/*
buypeeb - a program to track yahoo jp auctions of peebus.
Copyright (C) 2020 lynnesbian

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;
using System.Text.Json;
using System.Threading;
using System.Threading.Tasks;
using System.Net;
using System.Diagnostics;
using Gtk;

namespace Buypeeb {
	public struct ItemColumns {
		public const int Name = 0;
		public const int PriceYen = 1;
		public const int PriceAUD = 2;
		public const int Ending = 3;
		public const int Id = 4;
	}

	class MainWindow : Window {

		private string location;
		// private Queue<string> statuses;

		private ListStore items;
		private Settings settings;
		private TreeView itemTreeView;
		private Label statusLabel;
		private Builder builder;

		// TODO: whenever we get something from the builder, cache it for later
		// that way we don't need to constantly do "builder.GetObject"s
		// when that is done, you can use the cache array to replace everything from here...

		private Box selectionViewBox;
		private Label endingLabel;

		// ...to here.

		static SemaphoreSlim taskLimit = new SemaphoreSlim(4);

		private YahooAuctionsItem selectedItem {
			get {
				if (this.itemTreeView.Selection.CountSelectedRows() == 0) {
					// avoids incurring the wrath of Gtk-CRITICAL **
					return null;
				}

				this.itemTreeView.Selection.GetSelected(out TreeIter iter);
				return (YahooAuctionsItem)this.itemTreeView.Model.GetValue(iter, 0);
			}
		}

		public MainWindow() : this(new Builder("main.glade")) { }

		private MainWindow(Builder builder) : base(builder.GetObject("wndMain").Handle) {
			if (Environment.OSVersion.Platform == PlatformID.Win32NT) {
				// C:\Users\Beebus\AppData\Roaming\Lynnear Software\buypeeb
				this.location = System.IO.Path.Combine(Environment.ExpandEnvironmentVariables("%APPDATA%"), "Lynnear Software", "buypeeb");
			}
			else {
				// ~/.config/Lynnear Software/buypeeb
				this.location = System.IO.Path.Combine(Environment.GetEnvironmentVariable("HOME"), ".config", "Lynnear Software", "buypeeb");
			}

			string userdata = System.IO.Path.Combine(location, "userdata.json");
			if (File.Exists(userdata)) {
				try {
					string j = File.ReadAllText(userdata);
					this.settings = JsonSerializer.Deserialize<Settings>(j);
				}
				catch {
					// ???
					Console.WriteLine("oops");
					Application.Quit();
				}

			}
			else {
				this.settings = new Settings();
			}
			this.SaveSettings();
			this.Title = "Buypeeb";

			this.builder = builder;
			builder.Autoconnect(this);

			this.statusLabel = (Label)builder.GetObject("LabelStatus");
			this.selectionViewBox = (Box)builder.GetObject("SelectionViewBox");
			this.endingLabel = (Label)builder.GetObject("LabelSelectedEnding");

			// bind treeview columns to watchlist instead of needing to manually sync its liststore
			this.itemTreeView = (TreeView)builder.GetObject("TreeViewItems");
			this.items = new ListStore(typeof(YahooAuctionsItem));
			this.RenderList();

			this.itemTreeView.Model = this.items;
			TreeCellDataFunc[] funcs = {
				new TreeCellDataFunc(this.RenderColumnName),
				new TreeCellDataFunc(this.RenderColumnPriceYen),
				new TreeCellDataFunc(this.RenderColumnPriceAUD),
				new TreeCellDataFunc(this.RenderColumnEnding)
			};

			for (int i = 0; i < this.itemTreeView.Columns.Length; i++) {
				var c = this.itemTreeView.Columns[i];

				c.SetCellDataFunc(c.Cells[0], funcs[i]);
			}

			this.UpdateItems();
			GLib.Timeout.Add(1000, new GLib.TimeoutHandler(UpdateSelectionEndTime));

			DeleteEvent += WindowShutdown;
		}

		private void WindowShutdown(object sender, DeleteEventArgs args) {
			SaveSettings();
			Application.Quit();
		}

		// general behaviour

		private void SetStatus(string status) {
			this.statusLabel.Text = status ?? "Buypeeb";
		}

		private (TreePath path, TreeIter iter) GetRow(string id) {
			// TODO: surely there's a better way to do this
			TreeIter iter;
			this.itemTreeView.Model.GetIterFirst(out iter);

			for (int i = 0; i < this.itemTreeView.Model.IterNChildren(); i++) {
				var x = (YahooAuctionsItem)this.itemTreeView.Model.GetValue(iter, 0);
				if (x.id == id) {
					return (this.itemTreeView.Model.GetPath(iter), iter);
				}
				else {
					this.itemTreeView.Model.IterNext(ref iter);
				}
			}

			Console.WriteLine($"Couldn't find {id}!");
			return (null, iter);
		}

		private void SaveSettings() {
			string j = JsonSerializer.Serialize(this.settings);
			string p = System.IO.Path.Combine(this.location, "userdata.json");
			Console.WriteLine(j);
			if (!Directory.Exists(this.location)) {
				Directory.CreateDirectory(this.location);
			}
			if (!File.Exists(p)) {
				var fs = File.CreateText(p);
				fs.Close();
			}
			File.WriteAllText(System.IO.Path.Combine(this.location, "userdata.json"), j);
		}

		private void UpdateThread(string id) {
			var item = this.settings.watchlist[id];
			Console.WriteLine($"Updating {id}...");
			// set item.ready to false to show that it's still being updated
			// this changes a few behaviours, such as displaying the price as "..." instead of whatever's currently stored
			item.ready = false;

			Gtk.Application.Invoke(delegate {
				// TODO: find a way to not have to do this. i think we need to avoid actually modifying the items outside of the main thread :/
				var pathAndIter = this.GetRow(id);
				this.items.EmitRowChanged(pathAndIter.path, pathAndIter.iter);
			});

			using (WebClient client = new WebClient()) {
				// TODO: download should have timeout
				item.Update(client.DownloadString(item.url));
				// item.Update(File.ReadAllText("yahoo.html"));
			}

			Gtk.Application.Invoke(delegate {
				var pathAndIter = this.GetRow(id);
				this.items.EmitRowChanged(pathAndIter.path, pathAndIter.iter);
				if (item == this.selectedItem) {
					// if the user has this item selected and it just became ready, enable the selection box
					this.selectionViewBox.Sensitive = true;
				}
			});

			item.ready = true;
			Console.WriteLine($"{id} updated.");
		}

		private void UpdateItem(string id, bool force = false) {
			// don't start a new task if there are more than [tasklimit] tasks currently running
			// this makes sure we don't make 1000 simultaneous requests to yahoo auctions if there are 1000 items on the watchlist
			if (this.settings.watchlist[id].updatedRecently && !force) {
				// the item has been updated recently, and force is not true
				return;
			}

			taskLimit.Wait();
			var t = Task.Factory.StartNew(() => {
				this.UpdateThread(id);
			}).ContinueWith(task => { taskLimit.Release(); });
		}

		private void UpdateItems() {
			this.selectionViewBox.Sensitive = false;

			var t = Task.Factory.StartNew(() => {
				foreach (var item in this.settings.watchlist) {
					this.UpdateItem(item.Key);
				}
			});

		}

		private void UpdateSelectionView() {
			// get the currently selected item
			var item = this.selectedItem;
			var infobox = (Box)this.builder.GetObject("SelectionInfoBox");

			if (item == null) {
				this.selectionViewBox.Sensitive = false;
				infobox.Visible = false;
				var l = (Label)this.builder.GetObject("LabelSelectedName");
				l.Text = "buypeeb";
				return;
			}

			this.selectionViewBox.Sensitive = item.ready;
			infobox.Visible = true;

			var info = new Dictionary<string, string>();
			info.Add("Name", item.name);
			info.Add("YahooName", item.originalName);
			info.Add("Price", item.PriceJPY());
			info.Add("PriceAUD", item.PriceAUD());
			info.Add("Ending", "...");
			info.Add("Bids", $"{item.bids}");
			info.Add("BuyItNow", item.winPrice == 0 ? "No" : $"¥{item.PriceJPY(true)} (${item.PriceAUD(true)})");
			info.Add("AutoExtension", item.autoExtension ? "Yes" : "No");
			info.Add("LastUpdated", "Last updated: heeeenlo");

			foreach (var row in info) {
				var l = (Label)this.builder.GetObject($"LabelSelected{row.Key}");
				l.Text = row.Value;
			}

			var f = (ToggleButton)this.builder.GetObject("ButtonSelectedFavourite");
			f.Active = item.favourite;

		}
		private void OpenUrl(string url) {
			// https://github.com/dotnet/runtime/issues/17938
			if (Environment.OSVersion.Platform == PlatformID.Win32NT) {
				ProcessStartInfo psi = new ProcessStartInfo {
					FileName = url,
					UseShellExecute = true
				};
				Process.Start(psi);
			}
			else {
				// let's hope you have xdg-open installed
				Process.Start("xdg-open", url);
			}
		}

		private MessageDialog OkCancelDialogue(string message) {
			var md = new MessageDialog(
				parent_window: this,
				flags: DialogFlags.DestroyWithParent | DialogFlags.Modal,
				type: MessageType.Question,
				bt: ButtonsType.OkCancel,
				format: message
			);
			md.KeepAbove = true;
			md.Resizable = false;
			md.FocusOnMap = true;
			md.Title = "buypeeb";

			return md;
		}

		// show a simple entry dialogue that allows the user to enter text and either cancel or submit it
		private (Boolean accepted, string response) EntryDialogue(string title = "Buypeeb", string message = "Hi there!", string prefill = null) {
			Dialog ed = new Dialog(title, null, DialogFlags.DestroyWithParent | DialogFlags.Modal, "Cancel", ResponseType.Cancel, "OK", ResponseType.Ok);
			ed.DefaultResponse = ResponseType.Ok;
			ed.KeepAbove = true;

			Label edLabel = new Label(message);
			Entry edEntry = new Entry();
			if (!String.IsNullOrWhiteSpace(prefill)) {
				edEntry.Text = prefill;
			}
			edEntry.ActivatesDefault = true;

			ed.ContentArea.PackStart(edLabel, true, true, 2);
			edLabel.Show();
			ed.ContentArea.PackStart(edEntry, true, true, 10);
			edEntry.Show();

			ed.ContentArea.MarginBottom = 5;
			ed.ContentArea.MarginTop = 5;
			ed.ContentArea.MarginStart = 5;
			ed.ContentArea.MarginEnd = 5;
			ed.MarginBottom = 5;

			ResponseType accepted = (ResponseType)ed.Run();
			string response = edEntry.Text;
			ed.Dispose();
			return (accepted == ResponseType.Ok, response);
		}

		private void RenderList() {
			this.items.Clear();
			foreach (var item in this.settings.watchlist) {
				items.AppendValues(item.Value);
			}
		}

		// event handlers

		private void ButtonAddClicked(object sender, EventArgs a) {
			// Console.WriteLine("ButtonAddClicked");
			AddItemDialogue aid = new AddItemDialogue();
			aid.Title = "Buypeeb";
			ResponseType accepted = (ResponseType)aid.Run();
			string url = aid.GetURL();
			string name = aid.GetName();
			aid.Dispose();

			// vry simpl url validation for simpol creachers
			Regex rx = new Regex(@"^http.+yahoo.+");
			if (rx.IsMatch(url)) {
				Console.WriteLine("{0} will be added", url);
				this.UpdateItem(this.settings.Watch(url, name).id);
				this.RenderList();
			}
			else {
				Console.WriteLine("{0} is an invalid url", url);
			}
		}

		private void ButtonUpdateAllClicked(object sender, EventArgs a) {
			this.UpdateItems();
		}

		private void ButtonClearEndedClicked(object sender, EventArgs a) {
			Console.WriteLine("ButtonClearEndedClicked");
		}

		private void ButtonClearAllClicked(object sender, EventArgs a) {
			Console.WriteLine("ButtonClearAllClicked");
		}

		private void ButtonSaveClicked(object sender, EventArgs a) {
			this.SaveSettings();
		}

		private void ButtonQuitClicked(object sender, EventArgs a) {
			var md = this.OkCancelDialogue("Are you sure you want to quit?");

			ResponseType response = (ResponseType)md.Run();
			md.Dispose();
			if (response == ResponseType.Ok) {
				this.SaveSettings();
				Application.Quit();
			}
		}

		private void TreeViewItemsSelectionChanged(object sender, EventArgs a) {
			this.UpdateSelectionView();
		}

		private void ButtonViewBuyeeClicked(object sender, EventArgs a) {
			this.OpenUrl(this.selectedItem.buyeeUrl);
		}

		private void ButtonViewYahooClicked(object sender, EventArgs a) {
			this.OpenUrl(this.selectedItem.url);
		}

		private void ButtonSelectedRemoveClicked(object sender, EventArgs a) {
			var item = this.selectedItem;

			var md = this.OkCancelDialogue($"Are you sure you want to remove the item \"{item.name}\"?"); // TODO: this looks bad being all on one line

			ResponseType response = (ResponseType)md.Run();
			md.Dispose();
			if (response == ResponseType.Ok) {
				this.settings.watchlist.Remove(item.id);
				this.RenderList();
			}
		}

		private void ButtonSelectedRenameClicked(object sender, EventArgs a) {
			var item = this.selectedItem;
			(bool accepted, string response) = this.EntryDialogue("Rename item", $"Enter a new name for the item \"{item.name}\".", item.name);
			if (accepted) {
				item.name = response;
				this.UpdateSelectionView();
			}
		}

		private void ButtonSelectedUpdateClicked(object sender, EventArgs args) {
			this.selectionViewBox.Sensitive = false;
			this.UpdateItem(this.selectedItem.id);
		}

		private void ButtonSelectedFavouriteToggled(object sender, EventArgs args) {
			ToggleButton s = (ToggleButton)sender;
			this.selectedItem.favourite = s.Active;
		}

		// timers

		private bool UpdateSelectionEndTime() {
			if (!this.selectionViewBox.IsSensitive) {
				return true;
			}

			var item = this.selectedItem;
			if (!item.ready) {
				return true;
			}

			string ending = "";
			if (item.available) {
				var now = DateTime.Now;
				var end = item.endDate.ToLocalTime();
				var span = end.Subtract(now);

				if (span.Days > 0) {
					ending += span.ToString("dd' days, '"); // will format twelve days as "12 days, "
				}
				// timespan objects don't contain definitions for the time or date separators, so the colons need to be escaped
				// `HH` doesn't exist, but `hh` behaves identically
				// see https://docs.microsoft.com/en-us/dotnet/standard/base-types/standard-timespan-format-strings 
				ending += span.ToString(@"hh\:mm\:ss");
			}
			else {
				ending = "Auction has ended";
			}

			this.endingLabel.Text = ending;
			return true;
		}

		// column renderers

		private void RenderColumnName(Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.ITreeModel model, Gtk.TreeIter iter) {
			YahooAuctionsItem item = (YahooAuctionsItem)model.GetValue(iter, 0);
			(cell as Gtk.CellRendererText).Text = item.name ?? "Loading...";
		}

		private void RenderColumnPriceYen(Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.ITreeModel model, Gtk.TreeIter iter) {
			YahooAuctionsItem item = (YahooAuctionsItem)model.GetValue(iter, 0);
			(cell as Gtk.CellRendererText).Text = item.ready ? item.PriceJPY() : "...";
		}

		private void RenderColumnPriceAUD(Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.ITreeModel model, Gtk.TreeIter iter) {
			YahooAuctionsItem item = (YahooAuctionsItem)model.GetValue(iter, 0);
			(cell as Gtk.CellRendererText).Text = item.ready ? item.PriceAUD() : "...";
		}

		private void RenderColumnEnding(Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.ITreeModel model, Gtk.TreeIter iter) {
			YahooAuctionsItem item = (YahooAuctionsItem)model.GetValue(iter, 0);
			string ending = "";
			if (item.ready) {
				var now = DateTime.Now;
				var end = item.endDate.ToLocalTime();
				// TODO: should we show the year if the auction ends next year? 0uo
				if (end.DayOfYear != now.DayOfYear) {
					// the auction isn't ending today, so we should show the day it's ending on for clarity
					ending += end.ToString("MMM d ");
				}
				ending += end.ToString("HH:mm");
				if (this.settings.displaySecondsInList) {
					// add the seconds on to the end
					ending += end.ToString(":ss");
				}
			}
			(cell as Gtk.CellRendererText).Text = item.ready ? ending : "...";
		}
	}
}
